"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, List, Optional
from puffotter.flask.base import db
from otaku_info.enums import ListService, MediaType
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaItem import MediaItem


def load_service_ids(
        limit_media_type_to: Optional[MediaType] = None
) -> Dict[ListService, Dict[str, MediaId]]:
    """
    Loads Media IDs from the database and maps them to their service id
    :param limit_media_type_to: If set, limits the media type
    :return: The mapped ids
    """
    media_ids: List[MediaId] = MediaId.query.options(
        db.joinedload(MediaId.media_item)
          .subqueryload(MediaItem.media_ids)
    ).all()

    if limit_media_type_to is not None:
        media_ids = list(filter(
            lambda x: x.media_type == limit_media_type_to,
            media_ids
        ))

    mapped_ids: Dict[ListService, Dict[str, MediaId]] = {
        x: {} for x in ListService
    }
    for media_id in media_ids:
        mapped_ids[media_id.service][media_id.service_id] = media_id

    return mapped_ids



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import requests
from typing import List, Optional
from datetime import datetime
from bs4 import BeautifulSoup
from puffotter.flask.base import app
from otaku_info.external.entities.RedditLnRelease import RedditLnRelease


def load_ln_releases(year: Optional[int] = None) -> List[RedditLnRelease]:
    """
    Loads the light novel releases
    """
    releases: List[RedditLnRelease] = []
    current_year = datetime.utcnow().year

    if year is None:
        for year in range(2018, current_year + 2):
            releases += load_ln_releases(year)
        return releases

    tables = load_tables(year)

    for i, table in enumerate(tables):
        month_number = i + 1

        for entry in table.find_all("tr"):
            release = RedditLnRelease.from_parts(year, entry.find_all("td"))

            if month_number != release.release_date.month:
                app.logger.debug(
                    f"Incorrect month: "
                    f"{month_number} != {release.release_date.month} "
                    f"({release.release_date}/{release.series_name})"
                )

            releases.append(release)

    return releases


def load_tables(year: int) -> List[BeautifulSoup]:
    """
    Loads the tables containing the release data
    """
    current_year = datetime.utcnow().year

    # TODO Parse years from 2015-2017
    if year < 2018 or year > current_year + 1:
        return []

    if year >= current_year:
        url = "https://old.reddit.com/r/LightNovels/wiki/upcomingreleases"
    else:
        url = f"https://old.reddit.com/r/LightNovels/wiki/{year}releases"

    headers = {"User-Agent": "Mozilla/5.0"}
    resp = requests.get(url, headers=headers).text
    soup = BeautifulSoup(resp, "html.parser")

    tables = soup.find_all("tbody")

    # Table 0: Releases for current month on side bar
    # Table 1: Table below current month releases on side bar
    # Table -1: To be announced
    tables = tables[2:-1]

    if year >= current_year:
        if year == current_year:
            tables = tables[0:12]
        else:
            tables = tables[12:]

    return tables



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask import request, render_template, redirect, url_for
from flask.blueprints import Blueprint
from flask_login import login_required, current_user
from otaku_info.enums import ListService, MediaType, MediaSubType
from otaku_info.db.MediaList import MediaList
from otaku_info.wrappers.UpdateWrapper import UpdateWrapper


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)

    @blueprint.route("/updates", methods=["POST"])
    @login_required
    def redirect_updates():
        """
        Redirects a POST requests to the appropriate GET request for
        the /updates route
        :return: The response
        """
        service, media_type, list_name = \
            request.form["list_ident"].split(":", 2)
        mincount = request.form.get("mincount", "0")
        include_complete = request.form.get("include_complete", "off") == "on"
        filter_subtype = request.form.get("filter_subtype")

        get_url = url_for(
            "updates.show_updates",
            service=service,
            media_type=media_type,
            list_name=list_name,
            mincount=mincount,
            include_complete=1 if include_complete else 0,
            filter_subtype=filter_subtype
        )
        return redirect(get_url)

    @blueprint.route("/updates", methods=["GET"])
    @login_required
    def show_updates():
        """
        Shows the user's manga updates for a specified service and list
        :return: The response
        """
        service_name = request.args.get("service")
        media_type_name = request.args.get("media_type")
        list_name = request.args.get("list_name")
        mincount = int(request.args.get("mincount", "0"))
        include_complete = request.args.get("include_complete", "0") == "1"
        subtype_name = request.args.get("filter_subtype")

        if service_name is None \
                or list_name is None \
                or media_type_name is None:
            media_lists = [
                x for x in MediaList.query.filter_by(user=current_user)
            ]
            media_lists.sort(key=lambda x: x.name)
            media_lists.sort(key=lambda x: x.media_type.value)
            media_lists.sort(key=lambda x: x.service.value)
            return render_template(
                "generic/updates.html",
                media_lists=media_lists
            )
        else:
            service = ListService(service_name)
            media_type = MediaType(media_type_name)
            if subtype_name is None:
                subtype = None
            else:
                subtype = MediaSubType(subtype_name)

            updates = UpdateWrapper.from_db(
                current_user,
                list_name,
                service,
                media_type,
                subtype,
                mincount,
                include_complete
            )
            return render_template(
                "generic/updates.html",
                updates=updates,
                list_name=list_name,
                service=service,
                media_type=media_type
            )

    return blueprint



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import json
import requests
from typing import Optional
from otaku_info.external.entities.MangadexItem import MangadexItem


def fetch_mangadex_item(mangadex_id: int) -> Optional[MangadexItem]:
    """
    Fetches information for a mangadex
    """
    endpoint = "https://mangadex.org/api/manga/{}".format(mangadex_id)
    try:
        with requests.get(endpoint) as response:
            data = json.loads(response.text)
            if data["status"] == "OK":
                return MangadexItem.from_json(mangadex_id, data)
            else:
                return None
    except json.JSONDecodeError:
        return None



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional, Dict, Any, Tuple
from otaku_info.enums import MediaType, MediaSubType, \
    ReleasingState, MediaRelationType, ListService


class AnimeListItem:
    """
    Class that models a general anime list item
    """
    def __init__(
            self,
            _id: int,
            service: ListService,
            extra_ids: Dict[ListService, str],
            media_type: MediaType,
            media_subtype: MediaSubType,
            english_title: Optional[str],
            romaji_title: str,
            cover_url: str,
            chapters: Optional[int],
            volumes: Optional[int],
            episodes: Optional[int],
            releasing_state: ReleasingState,
            relations: Dict[Tuple[MediaType, int], MediaRelationType]
    ):
        """
        Initializes the AnimeListItem object
        :param _id: The anime list ID
        :param service: The anime list service
        :param extra_ids: Any extra IDs for other services
        :param media_type: The media type of the series
        :param media_subtype: The media subtype of the series
        :param english_title: The English title of the series
        :param romaji_title: The Japanes title of the series written in romaji
        :param cover_url: URL to a cover image for the series
        :param chapters: The total amount of known manga chapters
        :param volumes: The total amount of known manga/ln volumes
        :param episodes: The total amount of known anime episodes
        :param releasing_state: The current releasing state of the series
        :param relations: Related media items identified by IDs
        """
        self.id = _id
        self.service = service
        self.extra_ids = extra_ids
        self.media_type = media_type
        self.media_subtype = media_subtype
        self.english_title = english_title
        self.romaji_title = romaji_title
        self.cover_url = cover_url
        self.chapters = chapters
        self.volumes = volumes
        self.episodes = episodes
        self.releasing_state = releasing_state
        self.relations = relations

    @property
    def latest_release(self) -> Optional[int]:
        """
        :return: The latest release. Chapters for manga, episodes for anime
        """
        if self.media_type == MediaType.ANIME:
            return self.episodes
        else:
            return self.chapters

    @classmethod
    def from_query(cls, media_type: MediaType, data: Dict[str, Any]) \
            -> "AnimeListItem":
        """
        Generates an AnimeListItem from a dictionary generated by an APi query
        :param media_type: The media type of the item
        :param data: The data to use
        :return: The generated AnimeListItem
        """
        raise NotImplementedError()



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os


sentry_dsn = "https://f899b0c46d324f37b83527a3994afd8d@sentry.namibsun.net/18"
"""
The sentry DSN used for error logging
"""

root_path: str = os.path.join(os.path.dirname(os.path.abspath(__file__)))
"""
The root path of the application
"""



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import Dict, List
from puffotter.flask.base import db, app
from puffotter.flask.db.User import User
from otaku_info.enums import MediaType, MediaSubType
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.MediaNotification import MediaNotification
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.NotificationSetting import NotificationSetting
from otaku_info.wrappers.UpdateWrapper import UpdateWrapper


def send_new_update_notifications():
    """
    Sends out telegram notifications for media updates
    :return: None
    """
    start = time.time()
    app.logger.info("Starting check for notifications")

    user_states: List[MediaUserState] = MediaUserState.query\
        .options(
            db.joinedload(MediaUserState.media_id)
              .subqueryload(MediaId.media_item)
              .subqueryload(MediaItem.media_ids)
        ) \
        .options(
            db.joinedload(MediaUserState.media_id)
              .subqueryload(MediaId.chapter_guess)
        ) \
        .options(
            db.joinedload(MediaUserState.user)
              .subqueryload(User.telegram_chat_id)
        ) \
        .options(db.joinedload(MediaUserState.media_notification)) \
        .all()

    notification_settings: Dict[int, NotificationSetting] = {
        x.user_id: x
        for x in NotificationSetting.query.all()
    }

    for user_state in user_states:

        settings = notification_settings.get(user_state.user_id)
        if settings is None or not settings.value:
            continue
        else:
            handle_notification(user_state, settings)

    db.session.commit()
    app.logger.info(f"Completed check for notifications in "
                    f"{time.time() - start}s.")


def handle_notification(
        media_user_state: MediaUserState,
        settings: NotificationSetting
):
    """
    Handles a single notification
    :param media_user_state: The user state for which to notify
    :param settings: The notification settings
    :return: None
    """
    chat = media_user_state.user.telegram_chat_id
    if chat is None:
        return

    update = UpdateWrapper(media_user_state)
    notification = media_user_state.media_notification
    if notification is None:
        notification = MediaNotification(
            media_user_state=media_user_state, last_update=update.latest
        )
        db.session.add(notification)

    if update.diff <= 0:
        return

    if notification.last_update < update.latest:
        notification.last_update = update.latest

        if update.score is not None and update.score >= settings.minimum_score:

            media_item = media_user_state.media_id.media_item
            if media_item.media_type == MediaType.ANIME:
                keyword = "Episode"
            elif media_item.media_subtype == MediaSubType.NOVEL:
                keyword = "Volume"
            else:
                keyword = "Chapter"

            chat.send_message(
                f"New {keyword} for {update.title}\n\n"
                f"{keyword} {update.progress}/{update.latest} "
                f"(+{update.diff})\n\n"
                f"{update.url}"
            )



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any, Optional
from otaku_info.enums import ListService, ReleasingState
from otaku_info.mappings import mangadex_external_id_names, \
    list_service_id_types


class MangadexItem:
    """
    Class that models a general anilist list item
    Represents the information fetched using anilist's API
    """
    def __init__(
            self,
            mangadex_id: int,
            external_ids: Dict[ListService, str],
            title: str,
            cover_url: str,
            total_chapters: Optional[int],
            latest_chapter: Optional[int],
            releasing_state: ReleasingState
    ):
        """
        Initializes the MangadexItem object
        :param mangadex_id: The mangadex ID of the series
        :param external_ids: IDs for other services
        :param title: The title of the series
        :param cover_url: URL to a cover image for this series
        :param total_chapters: The total amount of chapters
        :param latest_chapter: The latest chapter
        :param releasing_state: The releasing state
        """
        self.mangadex_id = mangadex_id
        self.external_ids = external_ids
        self.title = title
        self.cover_url = cover_url
        self.total_chapters = total_chapters
        self.latest_chapter = latest_chapter
        self.releasing_state = releasing_state

    @classmethod
    def from_json(cls, mangadex_id: int, data: Dict[str, Any]) \
            -> "MangadexItem":
        """
        Generates a MangadexItem from a dictionary generated by an API query
        :param mangadex_id: The mangadex ID
        :param data: The data to use
        :return: The generated MangadexItem
        """
        ids = {ListService.MANGADEX: str(mangadex_id)}
        links = data["manga"]["links"]
        if links is not None:
            for service, identifier in mangadex_external_id_names.items():
                if identifier in links:
                    _id = links[identifier]
                    id_type = list_service_id_types[service]
                    if id_type == int:
                        _id = "".join([x for x in _id if x.isdigit()])
                    try:
                        ids[service] = str(id_type(_id))
                    except ValueError:
                        pass

        total_chapters = cls.parse_chapter(data["manga"]["last_chapter"])
        if total_chapters == 0:
            total_chapters = None

        english_chapters = [
            cls.parse_chapter(x["chapter"])
            for x in data.get("chapter", {}).values()
            if x["lang_code"] == "gb"
        ]
        english_chapters = [x for x in english_chapters if x is not None]
        english_chapters.sort(reverse=True)
        newest_chapter = None
        if len(english_chapters) > 0:
            newest_chapter = english_chapters[0]

        releasing_state = cls.resolve_releasing_state(data["manga"]["status"])

        return MangadexItem(
            mangadex_id,
            ids,
            data["manga"]["title"],
            data["manga"]["cover_url"],
            total_chapters,
            newest_chapter,
            releasing_state
        )

    @staticmethod
    def parse_chapter(chapter_string: str) -> Optional[int]:
        """
        Converts a mangadex chapter string into an integer
        :param chapter_string: The chapter string to convert
        :return: The converted integer
        """
        chapter_string = chapter_string.split(".")[0]
        chapter_string = "".join([
            x for x in chapter_string if x.isdigit()
        ])
        try:
            return int(chapter_string)
        except (ValueError, TypeError):
            return None

    @staticmethod
    def resolve_releasing_state(state_code: int) -> ReleasingState:
        """
        Translates mangadex status codes to releasing states
        :param state_code: The status code to translate
        :return: The releasing state
        """
        return {
            1: ReleasingState.RELEASING,
            2: ReleasingState.FINISHED,
            3: ReleasingState.CANCELLED,
            4: ReleasingState.UNKNOWN  # Hiatus
        }[state_code]



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import Dict, Optional, cast
from puffotter.flask.base import app, db
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaId import MediaId
from otaku_info.enums import ListService, MediaType
from otaku_info.utils.db.DbCache import DbCache
from otaku_info.external.anilist import load_anilist_info
from otaku_info.external.myanimelist import load_myanimelist_item
from otaku_info.external.mangadex import fetch_mangadex_item
from otaku_info.external.entities.AnimeListItem import AnimeListItem
from otaku_info.external.entities.MangadexItem import MangadexItem
from otaku_info.utils.db.updater import update_or_insert_item
from otaku_info.utils.db.convert import mangadex_item_to_media_item, \
    anime_list_item_to_media_item
from otaku_info.utils.db.load import load_service_ids
from otaku_info.mappings import list_service_priorities


def update_mangadex_data(
        start: int = 1,
        end: Optional[int] = None,
        refresh: bool = False
):
    """
    Goes through mangadex IDs sequentially and stores ID mappings for
    these entries if found.
    Stops once 100 consecutive entries didn't return any results
    :param start: Optionally specifies a starting index
    :param end: Optionally specifies an ending index
    :param refresh: If true, will update existing mangadex info
    :return: None
    """
    start_time = time.time()
    app.logger.info("Starting Mangadex Update")
    DbCache.cleanup()

    endcounter = 0
    mangadex_id = start - 1

    existing_ids = load_service_ids(MediaType.MANGA)
    while True:
        mangadex_id += 1

        if mangadex_id % 100 == 0:
            db.session.commit()
            DbCache.cleanup()

        if mangadex_id == end or endcounter > 100:
            break
        elif str(mangadex_id) in existing_ids[ListService.MANGADEX] \
                and not refresh:
            continue

        app.logger.debug(f"Probing mangadex id {mangadex_id}")
        mangadex_item = fetch_mangadex_item(mangadex_id)
        if mangadex_item is None:
            endcounter += 1
            app.logger.debug("Couldn't load mangadex info")
            continue
        else:
            endcounter = 0

        __update_database_with_mangadex_item(
            mangadex_item,
            existing_ids
        )

    db.session.commit()
    DbCache.cleanup()
    app.logger.info(f"Finished Mangadex Update in "
                    f"{time.time() - start_time}s.")


def __update_database_with_mangadex_item(
        mangadex_item: MangadexItem,
        existing_ids: Dict[ListService, Dict[str, MediaId]]
):
    """
    Updates the database with the contents of a single mangadex item
    :param mangadex_item: The mangadex item
    :param existing_ids: Existing IDs mapped to list services
    :return: None
    """
    media_item = __update_or_insert_mangadex_media_item(
        mangadex_item,
        existing_ids
    )
    media_id_mapping = media_item.media_id_mapping

    for service, service_id in mangadex_item.external_ids.items():
        if service not in media_id_mapping:
            media_id = MediaId(
                media_item_id=media_item.id,
                media_type=media_item.media_type,
                media_subtype=media_item.media_subtype,
                service=service,
                service_id=service_id
            )
            update_or_insert_item(media_id)


def __update_or_insert_mangadex_media_item(
        mangadex_item: MangadexItem,
        existing_ids: Dict[ListService, Dict[str, MediaId]]
) -> MediaItem:
    """
    Resolves the media item for a mangadex item
    Will create the media item and prune superfluous media items
    :param mangadex_item: The mangadex item
    :param existing_ids: Existing IDs mapped to list services
    :return: The media item
    """
    existing_mapped_ids = {}
    for service, service_id in mangadex_item.external_ids.items():
        media_id = existing_ids[service].get(service_id)
        if media_id is not None:
            existing_mapped_ids[media_id.service] = media_id

    media_item = None
    if len(existing_mapped_ids) == 0:

        for service in [ListService.ANILIST, ListService.MYANIMELIST]:
            if media_item is None and service in mangadex_item.external_ids:
                media_item = __load_anime_list_based_media_item(
                    mangadex_item, service
                )
        if media_item is None:
            media_item = mangadex_item_to_media_item(mangadex_item)
            media_item = cast(MediaItem, update_or_insert_item(media_item))
    else:
        # Ensure that all have the same media item
        for service in list_service_priorities:
            if service in existing_mapped_ids:
                media_item = existing_mapped_ids[service].media_item

        for service, media_id in existing_mapped_ids.items():
            if media_id.media_item != media_item:
                # Delete duplicate media items
                db.session.delete(media_id.media_item)
                db.session.commit()

    assert media_item is not None
    return media_item


def __load_anime_list_based_media_item(
        mangadex_item: MangadexItem,
        service: ListService
) -> Optional[MediaItem]:
    """
    Inserts or Updates a media item based on anilist data for a mangadex item
    :param mangadex_item: The mangadex item
    :return: The media item
    """
    service_id = mangadex_item.external_ids[service]

    service_item: Optional[AnimeListItem] = None
    if service == ListService.ANILIST:
        service_item = load_anilist_info(int(service_id), MediaType.MANGA)
    elif service == ListService.MYANIMELIST:
        service_item = load_myanimelist_item(int(service_id), MediaType.MANGA)

    if service_item is None:
        return None

    media_item = anime_list_item_to_media_item(service_item)
    media_item = cast(MediaItem, update_or_insert_item(media_item))
    return media_item



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Tuple
from puffotter.flask.db.ModelMixin import ModelMixin as PuffotterModelMixin


class ModelMixin(PuffotterModelMixin):
    """
    Class that define methods that greatly ease working with existing database
    entries
    """

    @property
    def identifier_tuple(self) -> Tuple:
        """
        :return: A tuple that's unique to this database entry
        """
        raise NotImplementedError()

    def update(self, new_data: "ModelMixin"):
        """
        Updates the data in this record based on another object
        :param new_data: The object from which to use the new values
        :return: None
        """
        raise NotImplementedError()



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import Dict, Tuple, cast, Optional
from puffotter.flask.base import app, db
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.LnRelease import LnRelease
from otaku_info.utils.db.DbCache import DbCache
from otaku_info.enums import MediaType, ListService
from otaku_info.external.reddit import load_ln_releases
from otaku_info.external.myanimelist import load_myanimelist_item
from otaku_info.external.anilist import load_anilist_info
from otaku_info.external.entities.AnimeListItem import AnimeListItem
from otaku_info.external.entities.RedditLnRelease import RedditLnRelease
from otaku_info.utils.db.updater import update_or_insert_item
from otaku_info.utils.db.convert import anime_list_item_to_media_item, \
    anime_list_item_to_media_id, ln_release_from_reddit_item


def update_ln_releases():
    """
    Updates the light novel releases
    :return: None
    """
    start = time.time()
    app.logger.info("Starting Light Novel Release Update")
    DbCache.cleanup()

    ln_releases = load_ln_releases()
    existing_releases: Dict[Tuple, LnRelease] = \
        cast(Dict[Tuple, LnRelease], DbCache.load_existing_items(LnRelease))

    series_media_items = {}
    for existing in existing_releases.values():
        key = existing.series_name
        if existing.media_item is not None:
            if key in series_media_items:
                other_id = series_media_items[key].media_item_id
                if existing.media_item_id != other_id:
                    app.logger.warning(f"Different Media Items for {key}")
            series_media_items[key] = existing

    for ln_release in ln_releases:
        release = ln_release_from_reddit_item(ln_release, None)
        identifier = release.identifier_tuple
        existing = existing_releases.get(identifier)

        if existing is None:
            media_item = create_media_item_from_reddit_ln_release(ln_release)
            media_item_id = None if media_item is None else media_item.id
            release.media_item_id = media_item_id
        elif existing.media_item_id is None:
            if release.series_name in series_media_items:
                release.media_item_id = \
                    series_media_items[release.series_name].media_item_id
            else:
                media_item = create_media_item_from_reddit_ln_release(
                    ln_release
                )
                media_item_id = None if media_item is None else media_item.id
                release.media_item_id = media_item_id
        else:
            release.media_item_id = existing.media_item_id

        update_or_insert_item(release)

    db.session.commit()
    DbCache.cleanup()
    app.logger.info(f"Finished Light Novel Release Update "
                    f"in {time.time() - start}s.")


def create_media_item_from_reddit_ln_release(
        ln_release: RedditLnRelease
) -> Optional[MediaItem]:
    """
    Creates a media item based on a reddit ln release
    Fills any missing entries in the database
    :param ln_release: The ln release
    :return: The media item
    """
    mal_id = ln_release.myanimelist_id
    anilist_id = ln_release.anilist_id

    # noinspection PyUnusedLocal
    info: Optional[AnimeListItem] = None
    if mal_id is None:
        return None
    elif anilist_id is not None:
        info = load_anilist_info(anilist_id, MediaType.MANGA)
    else:
        info = load_myanimelist_item(mal_id, MediaType.MANGA)

    if info is None:
        return None
    else:
        media_item = anime_list_item_to_media_item(info)
        media_item = cast(MediaItem, update_or_insert_item(media_item))
        media_id = anime_list_item_to_media_id(info, media_item)
        update_or_insert_item(media_id)
        if info.service == ListService.ANILIST:
            media_id = anime_list_item_to_media_id(
                info, media_item, ListService.MYANIMELIST
            )
            update_or_insert_item(media_id)

        return media_item



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import traceback
from sqlalchemy.exc import IntegrityError
from puffotter.flask.base import db, app
from otaku_info.utils.db.DbCache import DbCache
from otaku_info.db.ModelMixin import ModelMixin


def update_or_insert_item(
        to_add: ModelMixin,
        commit_updates: bool = False,
        reload_cache: bool = False
) -> ModelMixin:
    """
    Updates or creates an item that can be identified based on a unique tuple.
    Ensures that no duplicates are inserted across threads
    :param to_add: The item to add
    :param commit_updates: Whether or not to commit updates
    :param reload_cache: Reloads the cached data if set to True
    :return: None
    """
    existing_item = DbCache.get_existing_item(to_add)

    try:
        if existing_item is None:
            db.session.add(to_add)
            DbCache.update_item(to_add)
            db.session.commit()
        else:
            existing_item.update(to_add)
            if commit_updates:
                db.session.commit()

    except IntegrityError as e:
        if not reload_cache:
            app.logger.warning("Retrying Insert/Update")
            update_or_insert_item(to_add, commit_updates, True)
        else:
            app.logger.error(f"Failed to insert/update: {e}\n"
                             f"{traceback.format_exc()}")
            db.session.rollback()
            raise e

    retval = DbCache.get_existing_item(to_add)
    assert retval is not None
    return retval



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from datetime import datetime
from typing import Optional, List
from bs4.element import Tag
from otaku_info.utils.dates import map_month_name_to_month_number
from otaku_info.external.anilist import load_anilist_info
from otaku_info.enums import ListService, MediaType


class RedditLnRelease:
    """
    Object that acts as a wrapper around a light novel release on reddit.com
    """

    def __init__(
            self,
            series_name: str,
            year: int,
            release_date_string: str,
            volume: str,
            publisher: Optional[str],
            purchase_link: Optional[str],
            info_link: Optional[str],
            digital: bool,
            physical: bool
    ):
        """
        Initializes the object
        :param series_name: The name of the series
        :param year: The year of release
        :param release_date_string: The month and day of release
        :param volume: The volume number
        :param publisher: The publisher of the release
        :param purchase_link: A purchase link for the release
        :param info_link: Link to information for the release
        :param digital: Whether or not the release is digital
        :param physical: Whether or not the release is physical
        """
        self.series_name = series_name
        self.volume = volume
        self.publisher = publisher
        self.purchase_link = purchase_link
        self.info_link = info_link
        self.digital = digital
        self.physical = physical
        self.year = year
        self._release_date_string = release_date_string

    @property
    def release_date(self) -> datetime:
        """
        :return: The release date as a datetime object
        """
        month_name, day_string = self._release_date_string.split(" ")
        month = map_month_name_to_month_number(month_name)
        if month is None:
            month = 1
        try:
            day = int(day_string)
        except ValueError:
            day = 1
        return datetime(year=self.year, month=month, day=day)

    @property
    def release_date_string(self) -> str:
        """
        :return: The release date as a string (ISO 8601)
        """
        return self.release_date.strftime("%Y-%m-%d")

    @property
    def myanimelist_id(self) -> Optional[int]:
        """
        :return: The myanimelist ID, if available
        """
        if self.info_link is None or "myanimelist.net" not in self.info_link:
            return None

        url_parts = self.info_link.split("/")
        index = -1
        while not url_parts[index].isdigit():
            index -= 1
        return int(url_parts[index])

    @property
    def anilist_id(self) -> Optional[int]:
        """
        :return: The anilist ID, if available
        """
        if self.myanimelist_id is None:
            return None
        anilist_info = load_anilist_info(
            self.myanimelist_id, MediaType.MANGA, ListService.MYANIMELIST
        )
        if anilist_info is not None:
            return anilist_info.id
        else:
            return None

    @classmethod
    def from_parts(cls, year: int, parts: List[Tag]) -> "RedditLnRelease":
        """
        Generates a reddit LN release from BeautifulSoup td tags
        :param year: The year of the release
        :param parts: The td tags
        :return: The reddit ln release
        """
        purchase_link_item = parts[3].find("a")
        purchase_link = None
        if purchase_link_item is not None:
            purchase_link = purchase_link_item["href"]

        info_link_item = parts[1].find("a")
        info_link: Optional[str] = None
        if info_link_item is not None:
            info_link = info_link_item["href"]

        digital = "digital" in parts[4].text.lower()
        physical = "physical" in parts[4].text.lower()

        return cls(
            series_name=parts[1].text,
            year=year,
            release_date_string=parts[0].text,
            volume=parts[2].text,
            publisher=parts[3].text,
            purchase_link=purchase_link,
            info_link=info_link,
            digital=digital,
            physical=physical
        )



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
import json
import requests
from typing import Optional
from puffotter.flask.base import app
from otaku_info.enums import MediaType
from otaku_info.external.entities.MyanimelistItem import MyanimelistItem


def load_myanimelist_item(myanimelist_id: int, media_type: MediaType) \
        -> Optional[MyanimelistItem]:
    """
    Loads myanimelist data using the jikan API
    :param myanimelist_id: The myanimelist ID
    :param media_type: The media type
    :return: The myanimelist item
    """
    url = f"https://api.jikan.moe/v3/{media_type.value}/{myanimelist_id}"
    response = requests.get(url)

    if response.status_code == 503:
        # Sometimes jikan temporarily loses connection to myanimelist
        time.sleep(2)
        response = requests.get(url)
    if response.status_code < 300:
        return None

    data = json.loads(response.text)
    if data["type"] == "BadResponseException":
        return None
    elif data["type"] == "RateLimitException":
        time.sleep(30)
        app.logger.warning("Rate limited by jikan")
        return load_myanimelist_item(myanimelist_id, media_type)

    mal_item = MyanimelistItem.from_query(media_type, data)
    return mal_item



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any, Tuple
from puffotter.flask.base import db
from puffotter.flask.db.User import User
from puffotter.flask.db.ModelMixin import ModelMixin
from otaku_info.enums import ListService


class ServiceUsername(ModelMixin, db.Model):
    """
    Database model that stores an external service username for a user
    """

    __tablename__ = "service_usernames"
    """
    The name of the database table
    """

    __table_args__ = (
        db.UniqueConstraint(
            "user_id",
            "username",
            "service",
            name="unique_service_username"
        ),
    )
    """
    Makes sure that objects that should be unique are unique
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    user_id: int = db.Column(
        db.Integer,
        db.ForeignKey(
            "users.id", ondelete="CASCADE", onupdate="CASCADE"
        ),
        nullable=False
    )
    """
    The ID of the user associated with this service username
    """

    user: User = db.relationship(
        "User",
        backref=db.backref(
            "service_usernames", lazy=True, cascade="all,delete"
        )
    )
    """
    The user associated with this service username
    """

    username: str = db.Column(db.String(255), nullable=False)
    """
    The service username
    """

    service: ListService = db.Column(db.Enum(ListService), nullable=False)
    """
    The external service this item is a username for
    """

    @property
    def identifier_tuple(self) -> Tuple[int, str, ListService]:
        """
        :return: A tuple that uniquely identifies this database entry
        """
        return self.user_id, self.username, self.service

    def update(self, new_data: "ServiceUsername"):
        """
        Updates the data in this record based on another object
        :param new_data: The object from which to use the new values
        :return: None
        """
        self.user_id = new_data.user_id
        self.username = new_data.username
        self.service = new_data.service

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models
                                 will be included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "user_id": self.user_id,
            "username": self.username,
            "service": self.service.value
        }
        if include_children:
            data["user"] = self.user.__json__(include_children)
        return data



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional
from threading import get_ident
from typing import Dict, Tuple, Type
from puffotter.flask.base import app
from otaku_info.db.ModelMixin import ModelMixin


class DbCache:
    """
    Class that helps identifying existing items by caching all
    database entries
    """
    __cached: Dict[int, Dict[str, Dict[Tuple, ModelMixin]]] = {}
    """
    This dictionary stores the cached results for each class.
    Attention: This is mapped by class name, so database classes
    must all have unique names (which should be the case anyways).
    The first layer of the dictionary is mapped to thread IDs
    """

    @staticmethod
    def cleanup():
        """
        Removes any cached items for the current thread
        :return: None
        """
        thread_id = get_ident()
        if thread_id in DbCache.__cached:
            DbCache.__cached.pop(thread_id)

    @staticmethod
    def get_existing_item(item: ModelMixin) -> Optional[ModelMixin]:
        """
        Retrieves an existing item from the cache if it exists
        :param item: The item for which to retrieve the existing item
        :return: The existing item or None
        """
        identifier = item.identifier_tuple
        class_name = item.__class__
        class_key = item.__class__.__name__

        existing_items = DbCache.load_existing_items(class_name, False)
        existing_item = existing_items.get(identifier)

        needs_reload = False
        if existing_item is None:
            for other_existing in DbCache.__cached.values():
                others = other_existing.get(class_key, {})
                check = others.get(identifier)

                if check is not None:
                    needs_reload = True
                    break

        if needs_reload:
            existing_items = DbCache.load_existing_items(class_name, True)
            existing_item = existing_items.get(identifier)

        return existing_item

    @staticmethod
    def update_item(item: ModelMixin):
        """
        Updates a cached item
        :param item: The new item
        :return: None
        """
        existing_items = DbCache.load_existing_items(item.__class__, False)
        existing_items[item.identifier_tuple] = item

    @staticmethod
    def load_existing_items(cls: Type[ModelMixin], reload: bool = False) \
            -> Dict[Tuple, ModelMixin]:
        """
        Retrieves all existing items for a database class mapped to their
        identifier tuples.
        Automatically separates the results by thread.
        :param cls: The database class
        :param reload: Whether to force a reload of database data
        :return: The existing entries mapped to identifier tuples
        """
        thread_id = get_ident()
        class_name = cls.__name__

        if thread_id not in DbCache.__cached:
            DbCache.__cached[thread_id] = {}
        thread_cache = DbCache.__cached[thread_id]

        class_cache = thread_cache.get(class_name)
        if class_cache is None:
            thread_cache[class_name] = DbCache.__load(cls)
        elif reload:
            for key in list(class_cache.keys()):
                class_cache.pop(key)
            class_cache.update(DbCache.__load(cls))

        return thread_cache[class_name]

    @staticmethod
    def __load(cls: Type[ModelMixin]) -> Dict[Tuple, ModelMixin]:
        """
        Loads the data for a database class
        :param cls: The database class
        :return: The database data
        """
        app.logger.debug(f"Caching db content for {cls.__name__}")
        # noinspection PyUnresolvedReferences
        return {x.identifier_tuple: x for x in cls.query.all()}



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

# noinspection PyProtectedMember
from puffotter.flask.test.TestFramework import \
    _TestFramework as __TestFrameWork
from otaku_info.main import root_path
from otaku_info.Config import Config
from otaku_info.routes import blueprint_generators
from otaku_info.db import models


class _TestFramework(__TestFrameWork):
    """
    Class that models a testing framework for the flask application
    """
    module_name = "otaku_info"
    root_path = root_path
    config = Config
    models = models
    blueprint_generators = blueprint_generators



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any, Optional
from otaku_info.enums import MediaType, MediaSubType, \
    ReleasingState, MediaRelationType, ListService
from otaku_info.external.entities.AnimeListItem import AnimeListItem


class AnilistItem(AnimeListItem):
    """
    Class that models a general anilist list item
    Represents the information fetched using anilist's API
    """

    @property
    def myanimelist_id(self) -> Optional[int]:
        """
        :return: The myanimelist ID
        """
        mal_id = self.extra_ids.get(ListService.MYANIMELIST)
        if mal_id is not None and mal_id.isdigit():
            return int(mal_id)
        else:
            return None

    @classmethod
    def from_query(cls, media_type: MediaType, data: Dict[str, Any]) \
            -> "AnilistItem":
        """
        Generates an AnilistItem from a dictionary generated by an APi query
        :param media_type: The media type of the item
        :param data: The data to use
        :return: The generated AnilistItem
        """
        _media_subtype = data["format"]
        if _media_subtype is None:
            _media_subtype = "unknown"
        media_subtype = MediaSubType(_media_subtype.lower())

        _releasing_state = data["status"]
        if _releasing_state is None:
            _releasing_state = "unknown"
        releasing_state = ReleasingState(_releasing_state.lower())

        relations = {}
        for edge in data["relations"]["edges"]:
            node_id = edge["node"]["id"]
            node_type = MediaType(edge["node"]["type"].lower())
            relation_type = MediaRelationType(edge["relationType"].lower())
            relations[(node_type, node_id)] = relation_type

        extra_ids = {}
        mal_id = data["idMal"]
        if mal_id is not None:
            extra_ids[ListService.MYANIMELIST] = str(mal_id)

        return AnilistItem(
            data["id"],
            ListService.ANILIST,
            extra_ids,
            media_type,
            media_subtype,
            data["title"]["english"],
            data["title"]["romaji"],
            data["coverImage"]["large"],
            data["chapters"],
            data["volumes"],
            data["episodes"],
            releasing_state,
            relations
        )



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import List
from puffotter.flask.base import db, app
from otaku_info.db.MediaId import MediaId
from otaku_info.utils.db.DbCache import DbCache
from otaku_info.db.MangaChapterGuess import MangaChapterGuess
from otaku_info.enums import MediaType, ListService, MediaSubType


def update_manga_chapter_guesses():
    """
    Updates the manga chapter guesses
    :return: None
    """
    start = time.time()
    app.logger.info("Starting update of manga chapter guesses")
    DbCache.cleanup()

    anilist_ids: List[MediaId] = MediaId.query\
        .filter_by(
            media_type=MediaType.MANGA,
            media_subtype=MediaSubType.MANGA,
            service=ListService.ANILIST
        ) \
        .options(db.joinedload(MediaId.media_user_states)) \
        .options(db.joinedload(MediaId.media_item)) \
        .options(db.joinedload(MediaId.chapter_guess)) \
        .all()

    for anilist_id in anilist_ids:
        if len(anilist_id.media_user_states) == 0:
            # We only update chapter guesses for items that are actively used
            # by users, since each update requires an HTTP request to anilist.
            # We don't want to be rate limited.
            continue
        elif anilist_id.chapter_guess is None:
            new_guess = MangaChapterGuess(media_id=anilist_id)
            db.session.add(new_guess)
            db.session.commit()
            guess = new_guess
        else:
            guess = anilist_id.chapter_guess

        app.logger.debug(f"Updating chapter guess for {anilist_id.service_id}")
        guess.update_guess()
        db.session.commit()

    db.session.commit()
    DbCache.cleanup()
    app.logger.info(f"Finished updating manga chapter guesses "
                    f"in {time.time() - start}")



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from otaku_info.enums import MediaType, MediaSubType, \
    ReleasingState, ListService, MediaRelationType
from otaku_info.external.entities.AnimeListItem import AnimeListItem


class MyanimelistItem(AnimeListItem):
    """
    Class that models a general myanimelist list item
    Represents the information fetched using myanimelist's jikan API
    """

    @classmethod
    def from_query(cls, media_type: MediaType, data: Dict[str, Any]) \
            -> "MyanimelistItem":
        """
        Generates an MyanimelistItem from a dictionary generated
        by an APi query
        :param media_type: The media type of the item
        :param data: The data to use
        :return: The generated MyanimelistItem
        """
        media_subtype = cls.resolve_media_subtype(data["type"])
        releasing_state = cls.resolve_releasing_state(data["status"])

        relations = {}
        for edge_type, edges in data["related"].items():
            relation_type = cls.resolve_relation_type(edge_type)
            for edge in edges:
                node_type = MediaType(edge["type"])
                node_id = edge["mal_id"]
                relations[(node_type, node_id)] = relation_type

        return cls(
            data["mal_id"],
            ListService.MYANIMELIST,
            {},
            media_type,
            media_subtype,
            data["title_english"],
            data["title"],
            data["image_url"],
            data.get("chapters"),
            data.get("volumes"),
            data.get("episodes"),
            releasing_state,
            relations
        )

    @staticmethod
    def resolve_relation_type(relation_string: str) -> MediaRelationType:
        """
        Resolves myanimelist relation types
        :param relation_string: The string to resolve
        :return: The resolved MediaRelationType
        """
        base = {
            x.value.lower().replace("_", " "): x
            for x in MediaRelationType
        }
        base.update({
            "parent story": MediaRelationType.PARENT,
            "alternative setting": MediaRelationType.ALTERNATIVE
        })
        return base[relation_string.lower()]

    @staticmethod
    def resolve_releasing_state(releasing_string: str) -> ReleasingState:
        """
        Resolves myanimelist releasing state
        :param releasing_string: The string to resolve
        :return: The resolved ReleasingState
        """
        base = {
            x.value.lower().replace("_", " "): x
            for x in ReleasingState
        }
        base.update({
            "finished airing": ReleasingState.FINISHED,
            "publishing": ReleasingState.RELEASING,
            "airing": ReleasingState.RELEASING
        })
        return base[releasing_string.lower()]

    @staticmethod
    def resolve_media_subtype(subtype_string: str) -> MediaSubType:
        """
        Resolves myanimelist media subtype
        :param subtype_string: The string to resolve
        :return: The resolved MediaSubType
        """
        base = {
            x.value.lower().replace("_", " "): x
            for x in MediaSubType
        }
        base.update({
            "one-shot": MediaSubType.ONE_SHOT
        })
        return base[subtype_string.lower()]



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional
from puffotter.flask.db.User import User
from otaku_info.enums import ListService, MediaType, MediaSubType
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaList import MediaList
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.LnRelease import LnRelease
from otaku_info.external.entities.AnimeListItem import AnimeListItem
from otaku_info.external.entities.AnilistUserItem import AnilistUserItem
from otaku_info.external.entities.MangadexItem import MangadexItem
from otaku_info.external.entities.RedditLnRelease import RedditLnRelease


def anime_list_item_to_media_item(anime_list_item: AnimeListItem) -> MediaItem:
    """
    Converts an anime list item to a media item object
    :param anime_list_item: The anime list item
    :return: The media item
    """
    return MediaItem(
        media_type=anime_list_item.media_type,
        media_subtype=anime_list_item.media_subtype,
        english_title=anime_list_item.english_title,
        romaji_title=anime_list_item.romaji_title,
        cover_url=anime_list_item.cover_url,
        latest_release=anime_list_item.latest_release,
        latest_volume_release=anime_list_item.volumes,
        releasing_state=anime_list_item.releasing_state
    )


def mangadex_item_to_media_item(mangadex_item: MangadexItem) -> MediaItem:
    """
    Converts a mangadex item to a media item object
    :param mangadex_item: The mangadex item
    :return: The media item
    """
    return MediaItem(
        media_type=MediaType.MANGA,
        media_subtype=MediaSubType.MANGA,
        english_title=mangadex_item.title,
        romaji_title=mangadex_item.title,
        cover_url=mangadex_item.cover_url,
        latest_release=mangadex_item.total_chapters,
        latest_volume_release=None,
        releasing_state=mangadex_item.releasing_state
    )


def anime_list_item_to_media_id(
        anime_list_item: AnimeListItem,
        media_item: MediaItem,
        list_service: Optional[ListService] = None
) -> MediaId:
    """
    Converts an anime list item into a media id
    :param anime_list_item: The anime list item
    :param media_item: The corresponding media item
    :param list_service: Uses an extra ID with the specified list service
    :return: The generated media id
    """
    if list_service is not None:
        service_id = anime_list_item.extra_ids[list_service]
        service = ListService.MYANIMELIST
    else:
        service_id = str(anime_list_item.id)
        service = ListService.ANILIST

    return MediaId(
        media_item_id=media_item.id,
        service_id=str(service_id),
        service=service,
        media_type=media_item.media_type,
        media_subtype=media_item.media_subtype
    )


def anilist_user_item_to_media_user_state(
        anilist_user_item: AnilistUserItem,
        media_id: MediaId,
        user: User
) -> MediaUserState:
    """
    Generates a media user state based on an anilist user item
    :param anilist_user_item: The anilist user item
    :param media_id: The corresponding media id
    :param user: The corresponding user
    :return: The media user state
    """
    return MediaUserState(
        media_id_id=media_id.id,
        user_id=user.id,
        progress=anilist_user_item.progress,
        volume_progress=anilist_user_item.volume_progress,
        score=anilist_user_item.score,
        consuming_state=anilist_user_item.consuming_state
    )


def anilist_user_item_to_media_list(
        anilist_user_item: AnilistUserItem,
        user: User,
) -> MediaList:
    """
    Generates a media list from an anilist user item
    :param anilist_user_item: The anilist user item
    :param user: The corresponding user
    :return: The media list object
    """
    return MediaList(
        user_id=user.id,
        name=anilist_user_item.list_name,
        service=ListService.ANILIST,
        media_type=anilist_user_item.media_type
    )


def ln_release_from_reddit_item(
        ln_release: RedditLnRelease,
        media_item: Optional[MediaItem]
) -> LnRelease:
    """
    Creates an LnRelease object based on a RedditLnRelease object
    :param ln_release: The RedditLnRelease object
    :param media_item: An associated MediaItem
    :return: The LnRelease object
    """
    media_item_id = None if media_item is None else media_item.id
    return LnRelease(
        media_item_id=media_item_id,
        release_date_string=ln_release.release_date_string,
        series_name=ln_release.series_name,
        volume=ln_release.volume,
        publisher=ln_release.publisher,
        purchase_link=ln_release.purchase_link,
        digital=ln_release.digital,
        physical=ln_release.physical
    )



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import List, Optional, cast, Dict
from puffotter.flask.base import db, app
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaList import MediaList
from otaku_info.db.MediaListItem import MediaListItem
from otaku_info.db.MediaUserState import MediaUserState
from otaku_info.db.ServiceUsername import ServiceUsername
from otaku_info.external.anilist import load_anilist
from otaku_info.external.entities.AnilistUserItem import AnilistUserItem
from otaku_info.enums import ListService, MediaType
from otaku_info.utils.db.DbCache import DbCache
from otaku_info.utils.db.updater import update_or_insert_item
from otaku_info.utils.db.convert import anime_list_item_to_media_id, \
    anime_list_item_to_media_item, anilist_user_item_to_media_user_state, \
    anilist_user_item_to_media_list


# TODO Delete removed items from the database
# For example, user deletes a MediaList
# Or: user remove a media user state from his account
# Or: Anilist removes an entry
# Remark: Not really a priority tbh.
# Maybe do this in another, separate bg thread?
# -> Nope, then we don't have the anilist info
def update_anilist_data(usernames: Optional[List[ServiceUsername]] = None):
    """
    Retrieves all entries on the anilists of all users that provided
    an anilist username
    :param usernames: Can be used to override the usernames to use
    :return: None
    """
    start = time.time()
    app.logger.info("Starting Anilist Update")
    DbCache.cleanup()

    if usernames is None:
        usernames = ServiceUsername.query\
            .filter_by(service=ListService.ANILIST).all()

    anilist_data: Dict[
        ServiceUsername,
        Dict[MediaType, List[AnilistUserItem]]
    ] = {
        username: {
            media_type: load_anilist(username.username, media_type)
            for media_type in MediaType
        }
        for username in usernames
    }

    for username, anilist_info in anilist_data.items():
        for media_type, anilist_items in anilist_info.items():
            for anilist_item in anilist_items:
                __perform_update(
                    anilist_item,
                    username
                )

    db.session.commit()  # Commit pending Updates
    DbCache.cleanup()
    app.logger.info(f"Finished Anilist Update in {time.time() - start}s.")


def __perform_update(
        anilist_item: AnilistUserItem,
        username: ServiceUsername
):
    """
    Inserts or updates the contents of a single anilist user item
    :param anilist_item: The anilist user item
    :param username: The service username
    :return: None
    """
    media_item = anime_list_item_to_media_item(anilist_item)
    media_item = cast(MediaItem, update_or_insert_item(media_item))

    media_id = anime_list_item_to_media_id(anilist_item, media_item)
    media_id = cast(MediaId, update_or_insert_item(media_id))

    if anilist_item.myanimelist_id is not None:
        mal_id = anime_list_item_to_media_id(
            anilist_item, media_item, ListService.MYANIMELIST
        )
        update_or_insert_item(mal_id)

    user_state = anilist_user_item_to_media_user_state(
        anilist_item, media_id, username.user
    )
    user_state = cast(MediaUserState, update_or_insert_item(user_state))

    media_list = anilist_user_item_to_media_list(anilist_item, username.user)
    media_list = cast(MediaList, update_or_insert_item(media_list))

    media_list_item = MediaListItem(
        media_list_id=media_list.id,
        media_user_state_id=user_state.id
    )
    update_or_insert_item(media_list_item)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import List, Optional
from puffotter.flask.base import db
from puffotter.flask.db.User import User
from otaku_info.enums import MediaType, MediaSubType, ListService, \
    ReleasingState
from otaku_info.db.MediaList import MediaList
from otaku_info.db.MediaItem import MediaItem
from otaku_info.db.MediaId import MediaId
from otaku_info.db.MediaListItem import MediaListItem
from otaku_info.db.MediaUserState import MediaUserState


class UpdateWrapper:
    """
    Class that encapsulates important data to display for manga updates
    """

    def __init__(self, media_user_state: MediaUserState):
        """
        Initializes the MangaUpdate object
        :param media_user_state: The media user state contains everything we
                                 need. For performance reasons, it's assumed
                                 that all relations have been loaded already.
        """
        self.user_state = media_user_state
        self.title = self.user_state.media_id.media_item.title
        self.cover_url = self.user_state.media_id.media_item.cover_url
        self.url = self.user_state.media_id.media_item.own_url
        self.related_ids = self.user_state.media_id.media_item.media_ids
        self.related_ids.sort(key=lambda x: x.service.value)

        self.score = self.user_state.score
        self.progress = self.calculate_progress()
        self.latest = self.calculate_latest()

        if self.latest < self.progress:
            self.latest = self.progress

        self.diff = self.latest - self.progress

    def calculate_progress(self) -> int:
        """
        :return: The user's current progress
        """
        media_type = self.user_state.media_id.media_type
        subtype = self.user_state.media_id.media_item.media_subtype
        if media_type == MediaType.MANGA and subtype == MediaSubType.NOVEL:
            progress = self.user_state.volume_progress
        else:
            progress = self.user_state.progress
        if progress is None:
            progress = 0
        return progress

    def calculate_latest(self) -> int:
        """
        :return: The latest release number
        """
        media_type = self.user_state.media_id.media_type
        subtype = self.user_state.media_id.media_item.media_subtype
        if media_type == MediaType.MANGA and subtype == MediaSubType.NOVEL:
            latest = self.user_state.media_id.media_item.latest_volume_release
        elif media_type == MediaType.MANGA:
            chapter_guess = self.user_state.media_id.chapter_guess
            if chapter_guess is None:
                latest = self.user_state.media_id.media_item.latest_release
            else:
                latest = chapter_guess.guess
        else:
            latest = self.user_state.media_id.media_item.latest_release
        if latest is None:
            latest = 0
        return latest

    @classmethod
    def from_media_lists(
            cls,
            media_lists: List[MediaList],
            media_subtype: Optional[MediaSubType],
            minimum_diff: int,
            include_complete: bool
    ) -> List["UpdateWrapper"]:
        """
        Generates UpdateWrapper objects based on media lists
        To avoid huge performance problems, the relations should be fully
        loaded beforehand, preferably in a single query.
        Results can be filtered using the additional parameters
        :param media_lists: The media lists containing the items to wrap
        :param media_subtype: If specified, limits the results to a specific
                              media subtype (example: Light novels)
        :param minimum_diff: Specifies a minimum diff value
        :param include_complete: Specifies whether completed items should be
                                 included
        :return: A list of UpdateWrapper objects
        """
        updates = []
        for media_list in media_lists:
            for media_list_item in media_list.media_list_items:
                user_state = media_list_item.media_user_state
                media_item = user_state.media_id.media_item
                subtype = media_item.media_subtype
                state = media_item.releasing_state
                if not include_complete and state == ReleasingState.FINISHED:
                    continue
                if media_subtype is not None and media_subtype != subtype:
                    continue

                update = cls(user_state)

                if update.diff < minimum_diff:
                    continue

                updates.append(update)

        updates.sort(key=lambda x: x.score, reverse=True)
        return updates

    @classmethod
    def from_db(
            cls,
            user: User,
            list_name: str,
            service: ListService,
            media_type: MediaType,
            media_subtype: Optional[MediaSubType],
            minimum_diff: int,
            include_complete: bool
    ):
        """
        Generates UpdateWrapper objects based on a couple of parameters
        and the current database contents
        :param user: The user for whom to load the updates
        :param list_name: The list name for which to load the updates
        :param service: The service for which to load the updates
        :param media_type: The media type for which to load the updates
        :param media_subtype: If specified, limits the results to a specific
                              media subtype (example: Light novels)
        :param minimum_diff: Specifies a minimum diff value
        :param include_complete: Specifies whether completed items should be
                                 included
        :return: A list of UpdateWrapper objects
        """
        media_lists: List[MediaList] = MediaList.query \
            .filter_by(
                user=user,
                name=list_name,
                service=service,
                media_type=media_type
            ) \
            .options(
                db.joinedload(MediaList.media_list_items)
                  .subqueryload(MediaListItem.media_user_state)
                  .subqueryload(MediaUserState.media_id)
                  .subqueryload(MediaId.chapter_guess)
            ) \
            .options(
                db.joinedload(MediaList.media_list_items)
                  .subqueryload(MediaListItem.media_user_state)
                  .subqueryload(MediaUserState.media_id)
                  .subqueryload(MediaId.media_item)
                  .subqueryload(MediaItem.media_ids)
            ) \
            .options(
                db.joinedload(MediaList.media_list_items)
                  .subqueryload(MediaListItem.media_user_state)
                  .subqueryload(MediaUserState.media_id)
                  .subqueryload(MediaId.media_item)
                  .subqueryload(MediaItem.ln_releases)
            ) \
            .all()

        return cls.from_media_lists(
            media_lists, media_subtype, minimum_diff, include_complete
        )



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of otaku-info.

otaku-info is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

otaku-info is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with otaku-info.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Optional, Dict, Any, Tuple
from otaku_info.external.entities.AnilistItem import AnilistItem
from otaku_info.enums import MediaType, MediaSubType, \
    ReleasingState, ConsumingState, MediaRelationType, ListService


class AnilistUserItem(AnilistItem):
    """
    Class that models an anilist list item for a user
    Represents the information fetched using anilist's API
    """
    def __init__(
            self,
            _id: int,
            service: ListService,
            extra_ids: Dict[ListService, str],
            media_type: MediaType,
            media_subtype: MediaSubType,
            english_title: Optional[str],
            romaji_title: str,
            cover_url: str,
            chapters: Optional[int],
            volumes: Optional[int],
            episodes: Optional[int],
            releasing_state: ReleasingState,
            relations: Dict[Tuple[MediaType, int], MediaRelationType],
            score: Optional[int],
            progress: Optional[int],
            volume_progress: Optional[int],
            consuming_state: ConsumingState,
            list_name: str
    ):
        """
        Initializes the AnilistItem object
        :param _id: The anilist ID
        :param service: Anilist
        :param extra_ids: The myanimelist ID of the series
        :param media_type: The media type of the series
        :param media_subtype: The media subtype of the series
        :param english_title: The English title of the series
        :param romaji_title: The Japanes title of the series written in romaji
        :param cover_url: URL to a cover image for the series
        :param chapters: The total amount of known manga chapters
        :param volumes: The total amount of known manga/ln volumes
        :param episodes: The total amount of known anime episodes
        :param releasing_state: The current releasing state of the series
        :param relations: Related media items identified by IDs
        :param score: The user's score for the series
        :param progress: The user's progress for the series
        :param volume_progress: The user's volume progress
        :param consuming_state: The user's consumption state for the series
        :param list_name: Which of the user's lists this entry belongs to
        """
        super().__init__(
            _id,
            service,
            extra_ids,
            media_type,
            media_subtype,
            english_title,
            romaji_title,
            cover_url,
            chapters,
            volumes,
            episodes,
            releasing_state,
            relations
        )
        self.score = score
        self.progress = progress
        self.volume_progress = volume_progress
        self.consuming_state = consuming_state
        self.list_name = list_name

    @classmethod
    def from_query(
            cls,
            media_type: MediaType,
            data: Dict[str, Any]
    ) -> "AnilistUserItem":
        """
        Generates an AnilistUserItem from a dictionary generated
        by an APi query
        :param media_type: The media type of the item
        :param data: The data to use
        :return: The generated AnilistItem
        """
        base = AnilistItem.from_query(media_type, data["media"])
        consuming_state = ConsumingState(data["status"].lower())

        return cls(
            base.id,
            base.service,
            base.extra_ids,
            base.media_type,
            base.media_subtype,
            base.english_title,
            base.romaji_title,
            base.cover_url,
            base.chapters,
            base.volumes,
            base.episodes,
            base.releasing_state,
            base.relations,
            data["score"],
            data["progress"],
            data["progressVolumes"],
            consuming_state,
            data["list_name"]
        )
